@@ -25,6 +25,8 @@ static int machine_readable;		/* Generate machine-readable output */
 static int map_mode;			/* Bus mapping mode enabled */
 static int show_domains;		/* Show domain numbers (0=disabled, 1=auto-detected, 2=requested) */
 
+const char program_name[] = "lspci";
+
 static char options[] = "nvbxs:d:ti:mgMD" GENERIC_OPTIONS ;
 
 static char help_msg[] = "\
@@ -32,6 +34,7 @@ Usage: lspci [<switches>]\n\
 \n\
 -v\t\tBe verbose\n\
 -n\t\tShow numeric ID's\n\
+-nn\t\tShow both textual and numeric ID's (names & numbers)\n\
 -b\t\tBus-centric view (PCI addresses and IRQ's instead of those seen by the CPU)\n\
 -x\t\tShow hex-dump of the standard portion of config space\n\
 -xxx\t\tShow hex-dump of the whole config space (dangerous; root only)\n\
@@ -118,7 +121,11 @@ scan_device(struct pci_dev *p)
   d->present = xmalloc(64);
   memset(d->present, 1, 64);
   if (!pci_read_block(p, 0, d->config, 64))
-    die("Unable to read the standard configuration space header");
+    {
+      fprintf(stderr, "Unable to read the standard header from device "
+	      "%04x:%02x:%02x.%d\n", p->domain, p->bus, p->dev, p->func);
+      return NULL;
+    }
   if ((d->config[PCI_HEADER_TYPE] & 0x7f) == PCI_HEADER_TYPE_CARDBUS)
     {
       /* For cardbus bridges, we need to fetch 64 bytes more to get the
@@ -127,7 +134,7 @@ scan_device(struct pci_dev *p)
 	d->config_cached += 64;
     }
   pci_setup_cache(p, d->config, d->config_cached);
-  pci_fill_info(p, PCI_FILL_IDENT | PCI_FILL_IRQ | PCI_FILL_BASES | PCI_FILL_ROM_BASE | PCI_FILL_SIZES);
+  pci_fill_info(p, PCI_FILL_IDENT | PCI_FILL_CLASS | PCI_FILL_IRQ | PCI_FILL_BASES | PCI_FILL_ROM_BASE | PCI_FILL_SIZES);
   return d;
 }
 
@@ -253,13 +260,13 @@ show_terse(struct device *d)
 {
   int c;
   struct pci_dev *p = d->dev;
-  byte classbuf[128], devbuf[128];
+  char classbuf[128], devbuf[128];
 
   show_slot_name(d);
   printf(" %s: %s",
 	 pci_lookup_name(pacc, classbuf, sizeof(classbuf),
 			 PCI_LOOKUP_CLASS,
-			 get_conf_word(d, PCI_CLASS_DEVICE)),
+			 p->device_class),
 	 pci_lookup_name(pacc, devbuf, sizeof(devbuf),
 			 PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE,
 			 p->vendor_id, p->device_id));
@@ -271,7 +278,7 @@ show_terse(struct device *d)
       c = get_conf_byte(d, PCI_CLASS_PROG);
       x = pci_lookup_name(pacc, devbuf, sizeof(devbuf),
 			  PCI_LOOKUP_PROGIF | PCI_LOOKUP_NO_NUMBERS,
-			  get_conf_word(d, PCI_CLASS_DEVICE), c);
+			  p->device_class, c);
       if (c || x)
 	{
 	  printf(" (prog-if %02x", c);
@@ -497,13 +504,13 @@ show_pcix_nobridge(struct device *d, int where)
   static const byte max_outstanding[8] = { 1, 2, 3, 4, 8, 12, 16, 32 };
 
   printf("PCI-X non-bridge device\n");
-      
+
   if (verbose < 2)
     return;
 
   if (!config_fetch(d, where + PCI_PCIX_STATUS, 4))
     return;
-  
+
   command = get_conf_word(d, where + PCI_PCIX_COMMAND);
   status = get_conf_long(d, where + PCI_PCIX_STATUS);
   printf("\t\tCommand: DPERE%c ERO%c RBC=%d OST=%d\n",
@@ -531,18 +538,18 @@ show_pcix_nobridge(struct device *d, int where)
 static void
 show_pcix_bridge(struct device *d, int where)
 {
-  static const byte * const sec_clock_freq[8] = { "conv", "66MHz", "100MHz", "133MHz", "?4", "?5", "?6", "?7" };
+  static const char * const sec_clock_freq[8] = { "conv", "66MHz", "100MHz", "133MHz", "?4", "?5", "?6", "?7" };
   u16 secstatus;
   u32 status, upstcr, downstcr;
-  
+
   printf("PCI-X bridge device\n");
-  
+
   if (verbose < 2)
     return;
-  
+
   if (!config_fetch(d, where + PCI_PCIX_BRIDGE_STATUS, 12))
     return;
-  
+
   secstatus = get_conf_word(d, where + PCI_PCIX_BRIDGE_SEC_STATUS);
   printf("\t\tSecondary Status: 64bit%c 133MHz%c SCD%c USC%c SCO%c SRD%c Freq=%s\n",
 	 FLAG(secstatus, PCI_PCIX_BRIDGE_SEC_STATUS_64BIT),
@@ -553,7 +560,7 @@ show_pcix_bridge(struct device *d, int where)
 	 FLAG(secstatus, PCI_PCIX_BRIDGE_SEC_STATUS_SPLIT_REQUEST_DELAYED),
 	 sec_clock_freq[(secstatus >> 6) & 7]);
   status = get_conf_long(d, where + PCI_PCIX_BRIDGE_STATUS);
-  printf("\t\tStatus: Dev=%02x:%02x.%d 64bit%c 133MHz%c SCD%c USC%c SCO%c SRD%c\n", 
+  printf("\t\tStatus: Dev=%02x:%02x.%d 64bit%c 133MHz%c SCD%c USC%c SCO%c SRD%c\n",
 	 ((status >> 8) & 0xff),
 	 ((status >> 3) & 0x1f),
 	 (status & PCI_PCIX_BRIDGE_STATUS_FUNCTION),
@@ -988,7 +995,8 @@ show_msi(struct device *d, int where, int cap)
   u32 t;
   u16 w;
 
-  printf("Message Signalled Interrupts: 64bit%c Queue=%d/%d Enable%c\n",
+  printf("Message Signalled Interrupts: Mask%c 64bit%c Queue=%d/%d Enable%c\n",
+         FLAG(cap, PCI_MSI_FLAGS_MASK_BIT),
 	 FLAG(cap, PCI_MSI_FLAGS_64BIT),
 	 (cap & PCI_MSI_FLAGS_QSIZE) >> 4,
 	 (cap & PCI_MSI_FLAGS_QMASK) >> 1,
@@ -1009,6 +1017,26 @@ show_msi(struct device *d, int where, int cap)
     w = get_conf_word(d, where + PCI_MSI_DATA_32);
   t = get_conf_long(d, where + PCI_MSI_ADDRESS_LO);
   printf("%08x  Data: %04x\n", t, w);
+  if (cap & PCI_MSI_FLAGS_MASK_BIT)
+    {
+      u32 mask, pending;
+
+      if (is64)
+	{
+	  if (!config_fetch(d, where + PCI_MSI_MASK_BIT_64, 8))
+	    return;
+	  mask = get_conf_long(d, where + PCI_MSI_MASK_BIT_64);
+	  pending = get_conf_long(d, where + PCI_MSI_PENDING_64);
+	}
+      else
+        {
+	  if (!config_fetch(d, where + PCI_MSI_MASK_BIT_32, 8))
+	    return;
+	  mask = get_conf_long(d, where + PCI_MSI_MASK_BIT_32);
+	  pending = get_conf_long(d, where + PCI_MSI_PENDING_32);
+	}
+      printf("\t\tMasking: %08x  Pending: %08x\n", mask, pending);
+    }
 }
 
 static void show_vendor(void)
@@ -1064,9 +1092,9 @@ static void show_express_dev(struct device *d, int where, int type)
 
   w = get_conf_word(d, where + PCI_EXP_DEVCTL);
   printf("\t\tDevice: Errors: Correctable%c Non-Fatal%c Fatal%c Unsupported%c\n",
-	FLAG(w, PCI_EXP_DEVCTL_CERE), 
-	FLAG(w, PCI_EXP_DEVCTL_NFERE), 
-	FLAG(w, PCI_EXP_DEVCTL_FERE), 
+	FLAG(w, PCI_EXP_DEVCTL_CERE),
+	FLAG(w, PCI_EXP_DEVCTL_NFERE),
+	FLAG(w, PCI_EXP_DEVCTL_FERE),
 	FLAG(w, PCI_EXP_DEVCTL_URRE));
   printf("\t\tDevice: RlxdOrd%c ExtTag%c PhantFunc%c AuxPwr%c NoSnoop%c\n",
 	FLAG(w, PCI_EXP_DEVCTL_RELAXED),
@@ -1268,6 +1296,22 @@ show_slotid(int cap)
 	 chs);
 }
 
+static void
+show_ssvid(struct device *d, int where)
+{
+  u16 subsys_v, subsys_d;
+  char ssnamebuf[256];
+
+  if (!config_fetch(d, where, 8))
+    return;
+  subsys_v = get_conf_word(d, where + PCI_SSVID_VENDOR);
+  subsys_d = get_conf_word(d, where + PCI_SSVID_DEVICE);
+  printf("Subsystem: %s\n",
+	   pci_lookup_name(pacc, ssnamebuf, sizeof(ssnamebuf),
+			   PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE,
+			   d->dev->vendor_id, d->dev->device_id, subsys_v, subsys_d));
+}
+
 static void
 show_aer(struct device *d UNUSED, int where UNUSED)
 {
@@ -1393,6 +1437,9 @@ show_caps(struct device *d)
 	    case PCI_CAP_ID_DBG:
 	      show_debug();
 	      break;
+	    case PCI_CAP_ID_SSVID:
+	      show_ssvid(d, where);
+	      break;
 	    case PCI_CAP_ID_EXP:
 	      show_express(d, where, cap);
 	      can_have_ext_caps = 1;
@@ -1484,7 +1531,7 @@ show_htype1(struct device *d)
 		   get_conf_long(d, PCI_PREF_BASE_UPPER32),
 		   pref_base,
 		   get_conf_long(d, PCI_PREF_LIMIT_UPPER32),
-		   pref_limit);
+		   pref_limit + 0xfffff);
 	}
     }
 
@@ -1589,7 +1636,7 @@ show_verbose(struct device *d)
   struct pci_dev *p = d->dev;
   word status = get_conf_word(d, PCI_STATUS);
   word cmd = get_conf_word(d, PCI_COMMAND);
-  word class = get_conf_word(d, PCI_CLASS_DEVICE);
+  word class = p->device_class;
   byte bist = get_conf_byte(d, PCI_BIST);
   byte htype = get_conf_byte(d, PCI_HEADER_TYPE) & 0x7f;
   byte latency = get_conf_byte(d, PCI_LATENCY_TIMER);
@@ -1758,6 +1805,19 @@ show_hex_dump(struct device *d)
     }
 }
 
+static void
+print_shell_escaped(char *c)
+{
+  printf(" \"");
+  while (*c)
+    {
+      if (*c == '"' || *c == '\\')
+	putchar('\\');
+      putchar(*c++);
+    }
+  putchar('"');
+}
+
 static void
 show_machine(struct device *d)
 {
@@ -1783,11 +1843,11 @@ show_machine(struct device *d)
 
   if (verbose)
     {
-      printf("Device:\t");
+      printf((machine_readable >= 2) ? "Slot:\t" : "Device:\t");
       show_slot_name(d);
       putchar('\n');
       printf("Class:\t%s\n",
-	     pci_lookup_name(pacc, classbuf, sizeof(classbuf), PCI_LOOKUP_CLASS, get_conf_word(d, PCI_CLASS_DEVICE)));
+	     pci_lookup_name(pacc, classbuf, sizeof(classbuf), PCI_LOOKUP_CLASS, p->device_class));
       printf("Vendor:\t%s\n",
 	     pci_lookup_name(pacc, vendbuf, sizeof(vendbuf), PCI_LOOKUP_VENDOR, p->vendor_id, p->device_id));
       printf("Device:\t%s\n",
@@ -1807,21 +1867,18 @@ show_machine(struct device *d)
   else
     {
       show_slot_name(d);
-      printf(" \"%s\" \"%s\" \"%s\"",
-	     pci_lookup_name(pacc, classbuf, sizeof(classbuf), PCI_LOOKUP_CLASS,
-			     get_conf_word(d, PCI_CLASS_DEVICE)),
-	     pci_lookup_name(pacc, vendbuf, sizeof(vendbuf), PCI_LOOKUP_VENDOR,
-			     p->vendor_id, p->device_id),
-	     pci_lookup_name(pacc, devbuf, sizeof(devbuf), PCI_LOOKUP_DEVICE,
-			     p->vendor_id, p->device_id));
+      print_shell_escaped(pci_lookup_name(pacc, classbuf, sizeof(classbuf), PCI_LOOKUP_CLASS, p->device_class));
+      print_shell_escaped(pci_lookup_name(pacc, vendbuf, sizeof(vendbuf), PCI_LOOKUP_VENDOR, p->vendor_id, p->device_id));
+      print_shell_escaped(pci_lookup_name(pacc, devbuf, sizeof(devbuf), PCI_LOOKUP_DEVICE, p->vendor_id, p->device_id));
       if (c = get_conf_byte(d, PCI_REVISION_ID))
 	printf(" -r%02x", c);
       if (c = get_conf_byte(d, PCI_CLASS_PROG))
 	printf(" -p%02x", c);
       if (sv_id && sv_id != 0xffff)
-	printf(" \"%s\" \"%s\"",
-	       pci_lookup_name(pacc, svbuf, sizeof(svbuf), PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_VENDOR, sv_id),
-	       pci_lookup_name(pacc, sdbuf, sizeof(sdbuf), PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_DEVICE, p->vendor_id, p->device_id, sv_id, sd_id));
+	{
+	  print_shell_escaped(pci_lookup_name(pacc, svbuf, sizeof(svbuf), PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_VENDOR, sv_id));
+	  print_shell_escaped(pci_lookup_name(pacc, sdbuf, sizeof(sdbuf), PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_DEVICE, p->vendor_id, p->device_id, sv_id, sd_id));
+	}
       else
 	printf(" \"\" \"\"");
       putchar('\n');
@@ -1933,7 +1990,7 @@ grow_tree(void)
   last_br = &host_bridge.chain;
   for(d=first_dev; d; d=d->next)
     {
-      word class = get_conf_word(d, PCI_CLASS_DEVICE);
+      word class = d->dev->device_class;
       byte ht = get_conf_byte(d, PCI_HEADER_TYPE) & 0x7f;
       if (class == PCI_CLASS_BRIDGE_PCI &&
 	  (ht == PCI_HEADER_TYPE_BRIDGE || ht == PCI_HEADER_TYPE_CARDBUS))
@@ -1996,7 +2053,7 @@ grow_tree(void)
 }
 
 static void
-print_it(byte *line, byte *p)
+print_it(char *line, char *p)
 {
   *p++ = '\n';
   *p = 0;
@@ -2008,10 +2065,10 @@ print_it(byte *line, byte *p)
       *p = ' ';
 }
 
-static void show_tree_bridge(struct bridge *, byte *, byte *);
+static void show_tree_bridge(struct bridge *, char *, char *);
 
 static void
-show_tree_dev(struct device *d, byte *line, byte *p)
+show_tree_dev(struct device *d, char *line, char *p)
 {
   struct pci_dev *q = d->dev;
   struct bridge *b;
@@ -2037,7 +2094,7 @@ show_tree_dev(struct device *d, byte *line, byte *p)
 }
 
 static void
-show_tree_bus(struct bus *b, byte *line, byte *p)
+show_tree_bus(struct bus *b, char *line, char *p)
 {
   if (!b->first_dev)
     print_it(line, p);
@@ -2064,7 +2121,7 @@ show_tree_bus(struct bus *b, byte *line, byte *p)
 }
 
 static void
-show_tree_bridge(struct bridge *b, byte *line, byte *p)
+show_tree_bridge(struct bridge *b, char *line, char *p)
 {
   *p++ = '-';
   if (!b->first_bus->sibling)
@@ -2076,7 +2133,7 @@ show_tree_bridge(struct bridge *b, byte *line, byte *p)
   else
     {
       struct bus *u = b->first_bus;
-      byte *k;
+      char *k;
 
       while (u->sibling)
         {
@@ -2289,7 +2346,7 @@ main(int argc, char **argv)
     switch (i)
       {
       case 'n':
-	pacc->numeric_ids = 1;
+	pacc->numeric_ids++;
 	break;
       case 'v':
 	verbose++;
@@ -2313,7 +2370,7 @@ main(int argc, char **argv)
 	show_tree++;
 	break;
       case 'i':
-	pacc->id_file_name = optarg;
+        pci_set_name_list_path(pacc, optarg, 0);
 	break;
       case 'm':
 	machine_readable++;
